from meshgraph.Graph import Graph
from meshgraph.Node import Node


def _load_file_content(path):
    with open(path) as f:
        lines = [line for line in f]
        return lines


def _parse_vertex_line(line, graph):
    coordinates = line.split()
    new_node = Node(float(coordinates[1]),
                    float(coordinates[2]),
                    float(coordinates[3]))
    new_node.index = len(graph.nodes)
    graph.add_node(new_node)


def _get_vert_index_from_group(vert_group):
    str_index = vert_group.split('/')[0]
    return int(str_index)-1


def _get_vertices_from_face_line(line):
    split_line = line.split()
    verts_indexes = []
    for i in range(1, len(split_line)):
        verts_indexes.append(_get_vert_index_from_group(split_line[i]))
    return verts_indexes


def _link_mutual_neighbors(graph, a, b):
    graph.get_node(a).add_neighbor(graph.get_node(b))
#   undirected graph:
    graph.get_node(b).add_neighbor(graph.get_node(a))


def _solve_graph_neighbors(face_vertices_indexes, graph):
    for i in range(0, len(face_vertices_indexes) - 1):
        _link_mutual_neighbors(graph, face_vertices_indexes[i], face_vertices_indexes[i + 1])
#   link first and last (loop the face vertices):
    _link_mutual_neighbors(graph, face_vertices_indexes[0], face_vertices_indexes[-1])


def _parse_face_line(line, graph):
    face_vertices_indexes = _get_vertices_from_face_line(line)
    _solve_graph_neighbors(face_vertices_indexes, graph)


def _parse_line(line, graph):
    line = line.strip()
    if line.startswith('v '):
        _parse_vertex_line(line, graph)
    elif line.startswith('f '):
        _parse_face_line(line, graph)


def _fill_graph_from_lines(lines, graph):
    for line in lines:
        _parse_line(line, graph)
    print("Graph loaded, %s verts." % len(graph.nodes))
    

def load_graph_from_string(content):
    graph = Graph()
    _fill_graph_from_lines(content.splitlines(), graph)
    return graph


def load_graph_from_file(file_path):
    graph = Graph()
    lines = _load_file_content(file_path)
    _fill_graph_from_lines(lines, graph)
    return graph



from typing import List
from meshgraph.Node import Node


class Graph:
    """
    Undirected Graph

    ...

    Attributes
    ----------
    nodes : List[Node]
        nodes (or vertices) composing the graph

    Methods
    -------
    add_node(node)
        Adds new node to the graph
    subtract_node(sub_nodes)
        Subtracts a list of nodes from the existing nodes. It is not guaranteed that the remaining nodes are sorted
        by index. For this, use <subtract_nodes_sorted>
    subtract_nodes_sorted(sub_nodes)
        Subtracts a list of nodes from the existing nodes and sorts the remaining nodes by index
    reset_nodes_indexes()
        Resets the nodes' index. After calling this function, the first node will have index=0, the second one
        1 and so on...
    sort_nodes(sub_nodes)
        Sorts nodes by node index
    print()
        Prints all the nodes and their neighbors in a pretty format

    """

    __slots__ = ('nodes')

    def __init__(self):
        self.nodes = []

    def add_node(self, node: Node):
        self.nodes.append(node)

    def get_node(self, index: int) -> Node:
        return self.nodes[index]

    def subtract_nodes(self, sub_nodes: List[Node]):
        """Subtracts a list of nodes from the existing nodes. It is not guaranteed that the remaining nodes are sorted
        by index. For this, use <subtract_nodes_sorted>"""
        self.nodes = list(set(self.nodes) - set(sub_nodes))

    def subtract_nodes_sorted(self, sub_nodes: List[Node]):
        """Subtracts a list of nodes from the existing nodes and sorts the remaining nodes by index"""
        self.subtract_nodes(sub_nodes)
        self.sort_nodes()

    def reset_nodes_indexes(self):
        """Resets the nodes' index. After calling this function, the first node will have index=0, the second one
        1 and so on..."""
        if len(self.nodes) == 0:
            return
        current_index = 0
        for n in self.nodes:
            n.index = current_index
            current_index += 1

    def sort_nodes(self):
        """Sort nodes collection by node index"""
        self.nodes = sorted(self.nodes, key=lambda n: n.index)

    def print(self):
        """Prints a pretty string representation of the current node and its neighbors."""
        for n in self.nodes:
            print(n)
            print("\tneighbors:" + str(n.neighbors))



import math


class Vec3:
    """
    Vector 3D representing a point in Euclidian space

    ...

    Attributes
    ----------
    x : float
        the X coordinate
    y : float
        the Y coordinate
    z : float
        the Z coordinate

    Methods
    -------
    distance(other: Vec3)
        Returns the euclidean distance between the current vec3 and the other

    """

    __slots__ = ('x', 'y', 'z')

    def __init__(self, x: float, y: float, z: float):
        self.x = x
        self.y = y
        self.z = z

    def distance(self, other: 'Vec3') -> float:
        dx = (self.x - other.x)**2
        dy = (self.y - other.y)**2
        dz = (self.z - other.z)**2
        return math.sqrt(dx + dy + dz)

    def __eq__(self, other: 'Vec3'):
        return self.x == other.x \
               and self.y == other.y \
               and self.z == other.z

    def __str__(self):
        return "[" + str(self.x) \
            + "," + str(self.y) \
            + "," + str(self.z)\
            + "]"



from meshgraph.Vec3 import Vec3


class Node:
    """
    Node representing a vertex in Euclidian space

    ...

    Attributes
    ----------
    pos : Vec3
        The coordinates of the Node
    neighbors : Set[Node]
        The set containing the neighbors of the current node

    Methods
    -------
    add_neighbor(node: Node)
        Adds new node to the neighbors

    """

    __slots__ = ('pos', '_index', 'neighbors')

    def __init__(self, x, y, z, index=None):
        self.pos = Vec3(x, y, z)
        self._index = index
        self.neighbors = set()

    @property
    def index(self) -> int:
        return self._index

    @index.setter
    def index(self, value):
        self._index = value

    def add_neighbor(self, node: 'Node'):
        """Add new Node instance to neighbors"""
        self.neighbors.add(node)

    def __str__(self):
        return "V" + str(self._index) \
            + str(self.pos)

    def __repr__(self):
        return str(self)
